# -*- coding: utf-8 -*-

"""
Created on Mon Jan. 15, 2018

Description: This file contains a series of utilities designed to test the
features in the 'PredictionIntervals.py" package of the pymcmcstat module.  The
functions tested include:
    - empirical_quantiles(x, p = np.array([0.25, 0.5, 0.75])):

@author: prmiles
"""
from pymcmcstat.plotting.PredictionIntervals import PredictionIntervals
from pymcmcstat.utilities.progressbar import progress_bar
import test.general_functions as gf
import matplotlib.pyplot as plt
import unittest
from mock import patch
import numpy as np

# --------------------------------------------
class Observation_Sample_Test(unittest.TestCase):
    def test_does_observation_sample_unknown_sstype_cause_system_exit(self):
        PI = PredictionIntervals()
        s2elem = np.array([[2.0]])
        ypred = np.linspace(2.0, 3.0, num = 5)
        ypred = ypred.reshape(5,1)
        sstype = 4
        with self.assertRaises(SystemExit, msg = 'Unrecognized sstype'):
            PI._observation_sample(s2elem, ypred, sstype)

        sstype = 0
        opred = PI._observation_sample(s2elem, ypred, sstype)
        self.assertEqual(opred.shape, ypred.shape, msg = 'Shapes should match')
        
    def test_does_observation_sample_wrong_size_s2elem_break_right_size_array(self):
        PI = PredictionIntervals()
        s2elem = np.array([[2.0, 1.0]])
        ypred = np.linspace(2.0, 3.0, num = 5)
        ypred = ypred.reshape(5,1)
        sstype = 0
        with self.assertRaises(SystemExit, msg = 'Mismatched size for s2chain and model return'):
            PI._observation_sample(s2elem, ypred, sstype)
            
    def test_does_observation_sample_2_column_ypred_right_size_array(self):
        PI = PredictionIntervals()
        ypred = np.linspace(2.0, 3.0, num = 10)
        ypred = ypred.reshape(5,2)
        
        s2elem = np.array([[2.0, 1.0]])
        sstype = 0
        opred = PI._observation_sample(s2elem, ypred, sstype)
        self.assertEqual(opred.shape, ypred.shape, msg = 'Shapes are compatible')
        
        s2elem = np.array([[2.0]])
        opred = PI._observation_sample(s2elem, ypred, sstype)
        self.assertEqual(opred.shape, ypred.shape, msg = 'Shapes are compatible')
        
        sstype = 1
        opred = PI._observation_sample(s2elem, ypred, sstype)
        self.assertEqual(opred.shape, ypred.shape, msg = 'Shapes are compatible')
        
        sstype = 2
        opred = PI._observation_sample(s2elem, ypred, sstype)
        self.assertEqual(opred.shape, ypred.shape, msg = 'Shapes are compatible')
        
    def test_does_observation_sample_off_s2elem_greater_than_1_cause_system_exit(self):
        PI = PredictionIntervals()
        s2elem = np.array([[2.0, 1.0]])
        ypred = np.linspace(2.0, 3.0, num = 15)
        ypred = ypred.reshape(5,3)
        sstype = 0
        with self.assertRaises(SystemExit):
            PI._observation_sample(s2elem, ypred, sstype)

# --------------------------------------------
class AnalyzeDataStructure(unittest.TestCase):
    def test_basic_ds(self):
        PI = PredictionIntervals()
        DS = gf.basic_data_structure()
        ndatabatches, ncols = PI._analyze_data_structure(data = DS)
        self.assertEqual(ndatabatches, 1, msg = 'Expect 1 batch')
        self.assertEqual(ncols, [1], msg = 'Expect [1]')
        
    def test_ds_2_nbatch(self):
        PI = PredictionIntervals()
        DS = gf.non_basic_data_structure()
        ndatabatches, ncols = PI._analyze_data_structure(data = DS)
        self.assertEqual(ndatabatches, 2, msg = 'Expect 2 batches')
        self.assertEqual(ncols, [1, 2], msg = 'Expect [1, 1]')
        
    def test_basic_ds_shape_0(self):
        PI = PredictionIntervals()
        DS = gf.basic_data_structure()
        DS.shape = [(100,)] # remove column - code should add it back
        ndatabatches, ncols = PI._analyze_data_structure(data = DS)
        self.assertEqual(ndatabatches, 1, msg = 'Expect 1 batch')
        self.assertEqual(ncols, [1], msg = 'Expect [1]')
# --------------------------------------------
class SetupDataStructureForPrediction(unittest.TestCase):
    def test_basic_datapred(self):
        PI = PredictionIntervals()
        DS = gf.basic_data_structure()
        datapred = PI._setup_data_structure_for_prediction(data = DS, ndatabatches = 1)
        self.assertTrue(np.array_equal(datapred[0].xdata[0], DS.xdata[0]), msg = 'Arrays should match')
        self.assertTrue(np.array_equal(datapred[0].ydata[0], DS.ydata[0]), msg = 'Arrays should match')
        
    def test_non_basic_datapred(self):
        PI = PredictionIntervals()
        DS = gf.non_basic_data_structure()
        datapred = PI._setup_data_structure_for_prediction(data = DS, ndatabatches = 2)
        self.assertTrue(np.array_equal(datapred[0].xdata[0], DS.xdata[0]), msg = 'Arrays should match')
        self.assertTrue(np.array_equal(datapred[0].ydata[0], DS.ydata[0]), msg = 'Arrays should match')
        self.assertTrue(np.array_equal(datapred[1].xdata[0], DS.xdata[1]), msg = 'Arrays should match')
        self.assertTrue(np.array_equal(datapred[1].ydata[0], DS.ydata[1]), msg = 'Arrays should match')
# --------------------------------------------
class AssignFeaturesFromResultsStructure(unittest.TestCase):
    def check_dictionary(self, check_these, PID, results):
        for ct in check_these:
            self.assertTrue(np.array_equal(PID[str('_PredictionIntervals__{}'.format(ct))], results[ct]), msg = str('Arrays should match: {}'.format(ct)))
        self.assertTrue(np.array_equal(PID['_PredictionIntervals__nbatch'], results['model_settings']['nbatch']), msg = 'Arrays should match: nbatch')
        
    def test_feature_assignment(self):
        PI = PredictionIntervals()
        results = gf.setup_pseudo_results()
        PI._assign_features_from_results_structure(results = results)
        PID = PI.__dict__
        check_these = ['chain', 's2chain', 'parind', 'local', 'theta', 'sstype']
        self.check_dictionary(check_these, PID, results)
        
    def test_second_feature_assignment(self):
        PI = PredictionIntervals()
        results = gf.setup_pseudo_results()
        results = gf.removekey(results, 'sstype')
        PI._assign_features_from_results_structure(results = results)
        PID = PI.__dict__
        check_these = ['chain', 's2chain', 'parind', 'local', 'theta']
        self.check_dictionary(check_these, PID, results)
        self.assertEqual(PID['_PredictionIntervals__sstype'], 0, msg = 'Should default to 0')
        
# --------------------------------------------
class DetermineShapeOfReponse(unittest.TestCase):
    def test_basic_modelfunction(self):
        PI = PredictionIntervals()
        DS = gf.basic_data_structure()
        datapred = PI._setup_data_structure_for_prediction(data = DS, ndatabatches = 1)
        nrow, ncol = PI._determine_shape_of_response(modelfunction = gf.predmodelfun, ndatabatches = 1, datapred = datapred, theta = [3.0, 5.0])
        self.assertEqual(nrow, [100], msg = 'Expect [100]')
        self.assertEqual(ncol, [1], msg = 'Expect [1]')
        
    def test_basic_modelfunction_list_nbatch_2(self):
        PI = PredictionIntervals()
        DS = gf.non_basic_data_structure()
        datapred = PI._setup_data_structure_for_prediction(data = DS, ndatabatches = 2)
        nrow, ncol = PI._determine_shape_of_response(modelfunction = [gf.predmodelfun, gf.predmodelfun], ndatabatches = 2, datapred = datapred, theta = [3.0, 5.0])
        self.assertEqual(nrow, [100, 100], msg = 'Expect [100, 100]')
        self.assertEqual(ncol, [1, 1], msg = 'Expect [1, 1]')
        
    def test_non_basic_modelfunction(self):
        PI = PredictionIntervals()
        DS = gf.non_basic_data_structure()
        datapred = PI._setup_data_structure_for_prediction(data = DS, ndatabatches = 2)
        nrow, ncol = PI._determine_shape_of_response(modelfunction = gf.predmodelfun, ndatabatches = 2, datapred = datapred, theta = [3.0, 5.0])
        self.assertEqual(nrow, [100, 100], msg = 'Expect [100, 100]')
        self.assertEqual(ncol, [1, 1], msg = 'Expect [1, 1]')
 
# --------------------------------------------
class AnalyzeS2Chain(unittest.TestCase):
    def test_s2chain_index_n_eq_1(self):
        PI = PredictionIntervals()
        ndatabatches = 1
        s2chain = np.random.random_sample(size = (100,1))
        ncol = [1]
        s2chain_index = PI._analyze_s2chain(ndatabatches = ndatabatches, s2chain = s2chain, ncol = ncol)
        self.assertTrue(np.array_equal(s2chain_index, np.array([[0,1]])), msg = str('Arrays should match: {}'.format(s2chain_index)))
        
    def test_s2chain_index_n_neq_1_tc_eq_n(self):
        PI = PredictionIntervals()
        ndatabatches = 3
        s2chain = np.random.random_sample(size = (100,3))
        ncol = [1, 1, 1]
        s2chain_index = PI._analyze_s2chain(ndatabatches = ndatabatches, s2chain = s2chain, ncol = ncol)
        self.assertTrue(np.array_equal(s2chain_index, np.array([[0,1],[1,2],[2,3]])), msg = str('Arrays should match: {}'.format(s2chain_index)))
        
        ndatabatches = 3
        s2chain = np.random.random_sample(size = (100,4))
        ncol = [1, 2, 1]
        s2chain_index = PI._analyze_s2chain(ndatabatches = ndatabatches, s2chain = s2chain, ncol = ncol)
        self.assertTrue(np.array_equal(s2chain_index, np.array([[0,1],[1,3],[3,4]])), msg = str('Arrays should match: {}'.format(s2chain_index)))
        
    def test_s2chain_index_n_neq_1_tc_neq_n(self):
        PI = PredictionIntervals()
        ndatabatches = 3
        s2chain = np.random.random_sample(size = (100,3))
        ncol = [1, 2, 1]
        s2chain_index = PI._analyze_s2chain(ndatabatches = ndatabatches, s2chain = s2chain, ncol = ncol)
        self.assertTrue(np.array_equal(s2chain_index, np.array([[0,1],[1,2],[2,3]])), msg = str('Arrays should match: {}'.format(s2chain_index)))
        
    def test_s2chain_index_n_neq_1_tc_neq_n_neq_ndatabatches(self):
        PI = PredictionIntervals()
        ndatabatches = 5
        s2chain = np.random.random_sample(size = (100,3))
        ncol = [1, 2, 1]
        with self.assertRaises(SystemExit, msg = 'Unrecognized data structure'):
            PI._analyze_s2chain(ndatabatches = ndatabatches, s2chain = s2chain, ncol = ncol)

# --------------------------------------------
class SetupSstype(unittest.TestCase):
    def test_setup_sstype(self):
        PI = PredictionIntervals()
        PI._PredictionIntervals__sstype = 3
        sstype = PI._setup_sstype(sstype = None)
        self.assertEqual(sstype, PI._PredictionIntervals__sstype, msg = 'Expected 3')
        sstype = PI._setup_sstype(sstype = 2)
        self.assertEqual(sstype, 0, msg = 'Expected 0')

# --------------------------------------------
class CheckNsample(unittest.TestCase):
    def test_check_nsample(self):
        PI = PredictionIntervals()
        nsample = PI._check_nsample(nsample = 400, nsimu = 500)
        self.assertEqual(nsample, 400, msg = 'Expect 400')
        nsample = PI._check_nsample(nsample = None, nsimu = 500)
        self.assertEqual(nsample, 500, msg = 'Expect 500')
# --------------------------------------------
class DefineSamplePoints(unittest.TestCase):
    def test_define_sample_points_nsample_gt_nsimu(self):
        PI = PredictionIntervals()
        iisample, nsample = PI._define_sample_points(nsample = 1000, nsimu = 500)
        self.assertEqual(iisample, range(500), msg = 'Expect range(500)')
        self.assertEqual(nsample, 500, msg = 'Expect nsample updated to 500')
        
    @patch('numpy.random.rand')
    def test_define_sample_points_nsample_lte_nsimu(self, mock_rand):
        PI = PredictionIntervals()
        aa = np.random.rand([400,1])
        mock_rand.return_value = aa
        iisample, nsample = PI._define_sample_points(nsample = 400, nsimu = 500)
        self.assertTrue(np.array_equal(iisample, np.ceil(aa*500) - 1), msg = 'Expect range(500)')
        self.assertEqual(nsample, 400, msg = 'Expect nsample to stay 400')
        
# --------------------------------------------
class InitializePlotFeatures(unittest.TestCase):
    def test_initialize_plot_features(self):
        PI = PredictionIntervals()
        htmp, ax = PI._initialize_plot_features(ii = 0, jj = 0, ny = 1, figsizeinches = [10, 11])
        self.assertEqual(htmp.get_figwidth(), 10.0, msg = 'Figure width is 10in')
        self.assertEqual(htmp.get_figheight(), 11.0, msg = 'Figure height is 11in')
        self.assertEqual(htmp.get_label(), 'Batch # 0 | Column # 0', msg = 'Strings should match')
        self.assertEqual(ax.get_label(), '', msg = 'Expect empty string')
        
# --------------------------------------------
class AddBatchColumnTitle(unittest.TestCase):
    def test_add_batch_column_title_nbatch_gt_1(self):
        PI = PredictionIntervals()
        htmp, ax = PI._initialize_plot_features(ii = 0, jj = 0, ny = 1, figsizeinches = [10, 11])
        PI._add_batch_column_title(nbatch = 2, ny = 1, ii = 0, jj = 0)
        self.assertEqual(ax.get_title(), 'Batch #0, Column #0', msg = 'Strings should match')
        self.assertEqual(htmp.get_figwidth(), 10.0, msg = 'Expect 10.0')
        self.assertEqual(htmp.get_figheight(), 11.0, msg = 'Expect 11.0')
                         
    def test_add_batch_column_title_ny_gt_1(self):
        PI = PredictionIntervals()
        htmp, ax = PI._initialize_plot_features(ii = 0, jj = 0, ny = 1, figsizeinches = [10, 11])
        PI._add_batch_column_title(nbatch = 1, ny = 2, ii = 0, jj = 0)
        self.assertEqual(ax.get_title(), 'Column #0', msg = 'Strings should match')
        self.assertEqual(htmp.get_figwidth(), 10.0, msg = 'Expect 10.0')
        self.assertEqual(htmp.get_figheight(), 11.0, msg = 'Expect 11.0')
                         
# --------------------------------------------
class SetupLabels(unittest.TestCase):
    def test_setup_labels_pi_none_nlines_2(self):
        PI = PredictionIntervals()
        clabels, plabels = PI._setup_labels(prediction_intervals = None, nlines = 2)
        self.assertEqual(clabels, ['99% CI', '95% CI', '90% CI', '50% CI'], msg = 'String should match')
        self.assertEqual(plabels, ['95% PI'], msg = 'String should match')
        
    def test_setup_labels_pi_not_none_nlines_2(self):
        PI = PredictionIntervals()
        clabels, plabels = PI._setup_labels(prediction_intervals = 0, nlines = 2)
        self.assertEqual(clabels, ['95% CI'], msg = 'String should match')
        self.assertEqual(plabels, ['95% PI'], msg = 'String should match')
        
    def test_setup_labels_pi_none_nlines_1(self):
        PI = PredictionIntervals()
        clabels, plabels = PI._setup_labels(prediction_intervals = None, nlines = 1)
        self.assertEqual(clabels, ['95% CI'], msg = 'String should match')
        self.assertEqual(plabels, ['95% PI'], msg = 'String should match')
        
# --------------------------------------------
class CheckPIFlag(unittest.TestCase):
    def test_check_pi_flag_true_0(self):
        PI = PredictionIntervals()
        prediction_intervals = PI._check_prediction_interval_flag(plot_pred_int = True, prediction_intervals = 0)
        self.assertEqual(prediction_intervals, 0, msg = 'Expect 0')
        
    def test_check_pi_flag_true_none(self):
        PI = PredictionIntervals()
        prediction_intervals = PI._check_prediction_interval_flag(plot_pred_int = True, prediction_intervals = None)
        self.assertEqual(prediction_intervals, None, msg = 'Expect None')
        
    def test_check_pi_flag_false_0(self):
        PI = PredictionIntervals()
        prediction_intervals = PI._check_prediction_interval_flag(plot_pred_int = False, prediction_intervals = 0)
        self.assertEqual(prediction_intervals, None, msg = 'Expect None')
        
    def test_check_pi_flag_false_none(self):
        PI = PredictionIntervals()
        prediction_intervals = PI._check_prediction_interval_flag(plot_pred_int = False, prediction_intervals = None)
        self.assertEqual(prediction_intervals, None, msg = 'Expect None')
        
        
# --------------------------------------------
class SetupIntervalLimits(unittest.TestCase):
    def test_setup_int_lims_s2chain_none(self):
        PI = PredictionIntervals()
        lims = PI._setup_interval_limits(s2chain = None)
        self.assertTrue(np.array_equal(lims, np.array([0.005,0.025,0.05,0.25,0.5,0.75,0.9,0.975,0.995])), msg = str('Arrays should match: {}'.format(lims)))
        
    def test_setup_int_lims_s2chain_not_none(self):
        PI = PredictionIntervals()
        lims = PI._setup_interval_limits(s2chain = 1)
        self.assertTrue(np.array_equal(lims, np.array([0.025, 0.5, 0.975])), msg = str('Arrays should match: {}'.format(lims)))
        
# --------------------------------------------
class SetupCountingMetrics(unittest.TestCase):
    def test_setup_counting_metrics(self):
        PI = PredictionIntervals()
        ci = gf.setup_pseudo_ci()
        nbatch, nn, nlines = PI._setup_counting_metrics(credible_intervals = ci)
        self.assertEqual(nbatch, 1, msg = 'Expect nbatch = 1')
        self.assertEqual(nn, 2, msg = 'Expect nn = 2')
        self.assertEqual(nlines, 1, msg = 'Expect nlines = 1')

# --------------------------------------------
class SetupIntervalPlotting(unittest.TestCase):
    def check_these_settings(self, infigsizeinches, expected):
        PI = PredictionIntervals()
        ci = gf.setup_pseudo_ci()
        prediction_intervals, figsizeinches, nbatch, nn, clabels, plabels = PI._setup_interval_plotting(plot_pred_int = True, prediction_intervals = None, credible_intervals = ci, figsizeinches = infigsizeinches)
        self.assertEqual(figsizeinches, expected, msg = str('Expect {}'.format(expected)))
        self.assertEqual(prediction_intervals, None, msg = 'Expect None')
        self.assertEqual(nbatch, 1, msg = 'Expect nbatch = 1')
        self.assertEqual(nn, 2, msg = 'Expect nn = 2')
        self.assertEqual(clabels, ['95% CI'], msg = 'String should match')
        self.assertEqual(plabels, ['95% PI'], msg = 'String should match')
        
    def test_setup_interval_plotting(self):
        self.check_these_settings(infigsizeinches = [10, 11], expected = [10, 11])
        self.check_these_settings(infigsizeinches = None, expected = [7,5])

# --------------------------------------------
class SetupGenerationRequirements(unittest.TestCase):
    @classmethod
    def setup_generation(cls):
        aa = np.random.rand([400,1])
        PI = PredictionIntervals()
        results = gf.setup_pseudo_results()
        PI._assign_features_from_results_structure(results = results)
        return PI, aa, results
        
    def common_checks(self, results, chain, sstype, nsample, iisample):
        self.assertTrue(np.array_equal(chain, results['chain']), msg = str('Arrays should match: {}'.format(chain)))
        self.assertEqual(sstype, 0, msg = 'Expect 0')
        self.assertEqual(nsample, 100, msg = 'Expect nsample to go to 100')
        self.assertTrue(np.array_equal(iisample, range(100)), msg = str('Arrays should match: {}'.format(iisample)))

    @patch('numpy.random.rand')
    def test_setup_generation(self, mock_rand):
        PI, aa, results = self.setup_generation()
        mock_rand.return_value = aa
        chain, s2chain, lims, sstype, nsample, iisample = PI._setup_generation_requirements(nsample = 400, calc_pred_int = False, sstype = 0)
        
        self.assertTrue(np.array_equal(lims, np.array([0.005,0.025,0.05,0.25,0.5,0.75,0.9,0.975,0.995])), msg = str('Arrays should match: {}'.format(lims)))
        self.assertEqual(s2chain, None, msg = 'Expect None')

        self.common_checks(results, chain, sstype, nsample, iisample)
        
    @patch('numpy.random.rand')
    def test_setup_generation_with_pi(self, mock_rand):
        PI, aa, results = self.setup_generation()
        mock_rand.return_value = aa
        chain, s2chain, lims, sstype, nsample, iisample = PI._setup_generation_requirements(nsample = 400, calc_pred_int = True, sstype = 0)

        self.assertTrue(np.array_equal(lims, np.array([0.025, 0.5, 0.975])), msg = str('Arrays should match: {}'.format(lims)))
        self.assertTrue(np.array_equal(s2chain, results['s2chain']), msg = str('Arrays should match: {}'.format(s2chain)))
        
        self.common_checks(results, chain, sstype, nsample, iisample)
        
# --------------------------------------------
class SetupIntervalii(unittest.TestCase):
    @classmethod
    def setup_interval(cls):
        PI = PredictionIntervals()
        DS = gf.basic_data_structure()
        datapred = PI._setup_data_structure_for_prediction(data = DS, ndatabatches = 1)
        return PI, datapred

    def common_checks(self, datapredii, datapred, nrow, ncol, modelfun, predmodelfun, test):
        self.assertTrue(np.array_equal(datapredii.xdata[0], datapred[0].xdata[0]), msg = 'Arrays should match')
        self.assertTrue(np.array_equal(datapredii.ydata[0], datapred[0].ydata[0]), msg = 'Arrays should match')
        self.assertEqual(nrow, 100, msg = 'Expect nrow = 100')
        self.assertEqual(ncol, 1, msg = 'Expect ncol = 1')
        self.assertEqual(modelfun, predmodelfun, msg = 'Functions should match')
        self.assertTrue(np.array_equal(test, np.array([True, True])), msg = str('Arrays should match: {}'.format(test)))
        
    def test_setup_predii_nonlist_mdlfun(self):
        PI, datapred = self.setup_interval()
        datapredii, nrow, ncol, modelfun, test = PI._setup_interval_ii(ii = 0, datapred = datapred, nrow = [100], ncol = [1], modelfunction = gf.predmodelfun, local = np.array([0, 0]))
        self.common_checks(datapredii, datapred, nrow, ncol, modelfun, gf.predmodelfun, test)
        
    def test_setup_predii_list_mdlfun(self):
        PI, datapred = self.setup_interval()
        datapredii, nrow, ncol, modelfun, test = PI._setup_interval_ii(ii = 0, datapred = datapred, nrow = [100], ncol = [1], modelfunction = [gf.predmodelfun], local = np.array([0, 0]))
        self.common_checks(datapredii, datapred, nrow, ncol, modelfun, gf.predmodelfun, test)
        
# --------------------------------------------
def gq_settings():
    PI = PredictionIntervals()
    nsample = 500
    nrow = 100
    ncol = 1
    return PI, nsample, nrow, ncol

class GenerateQuantiles(unittest.TestCase):
    def test_generate_quantiles(self):
        PI, nsample, nrow, ncol = gq_settings()
        ysave = np.zeros([nsample, nrow, ncol])
        lims = np.array([0.025, 0.5, 0.975])
        cq = PI._generate_quantiles(response = ysave, lims = lims, ncol = ncol)
        self.assertTrue(isinstance(cq, list), msg = 'Expect list')
        self.assertEqual(cq[0].shape, (3,100), msg = 'Expect shape = (3, 100)')
        
    def test_generate_quantiles_s2chain_not_none(self):
        PI, nsample, nrow, ncol = gq_settings()
        ysave = np.zeros([nsample, nrow, ncol])
        osave = np.zeros([nsample, nrow, ncol])
        lims = np.array([0.025, 0.5, 0.975])
        cq = PI._generate_quantiles(response = ysave, lims = lims, ncol = ncol)
        pq = PI._generate_quantiles(response = osave, lims = lims, ncol = ncol)
        self.assertTrue(isinstance(cq, list), msg = 'Expect list')
        self.assertTrue(isinstance(pq, list), msg = 'Expect list')
        self.assertEqual(cq[0].shape, (3,100), msg = 'Expect shape = (3, 100)')
        self.assertEqual(pq[0].shape, (3,100), msg = 'Expect shape = (3, 100)')

# --------------------------------------------
def cc_setup(calc_pred_int = True):
    PI = PredictionIntervals()
    results = gf.setup_pseudo_results()
    PI._assign_features_from_results_structure(results = results)
    testchain, s2chain, lims, sstype, nsample, iisample = PI._setup_generation_requirements(nsample = 400, calc_pred_int = calc_pred_int, sstype = 0)
    DS = gf.basic_data_structure()
    datapred = PI._setup_data_structure_for_prediction(data = DS, ndatabatches = 1)
    return PI, results, testchain, s2chain, lims, sstype, nsample, iisample, datapred

class CalcCredii(unittest.TestCase):
    def test_calc_credii(self):
        PI, __, testchain, __, lims, sstype, nsample, iisample, datapred = cc_setup()
        
        ysave = PI._calc_credible_ii(testchain = testchain, nrow = 100, ncol = 1, waitbar = False, test = np.array([True, True]), modelfun = gf.predmodelfun, datapredii = datapred[0])
        self.assertTrue(isinstance(ysave, np.ndarray), msg = 'Expect array')
        self.assertEqual(ysave.shape[0], 100, msg = 'Expect 1st dim = 100')
        
    def test_calc_credii_with_waitbar(self):
        PI, __, testchain, __, lims, sstype, nsample, iisample, datapred = cc_setup()
        
        PI._PredictionIntervals__wbarstatus = progress_bar(iters = 200)
            
        ysave = PI._calc_credible_ii(testchain = testchain, nrow = 100, ncol = 1, waitbar = True, test = np.array([True, True]), modelfun = gf.predmodelfun, datapredii = datapred[0])
        self.assertTrue(isinstance(ysave, np.ndarray), msg = 'Expect array')
        self.assertEqual(ysave.shape[0], 100, msg = 'Expect 1st dim = 100')
        self.assertEqual(PI._PredictionIntervals__wbarstatus.percentage(3), 1.5, msg = 'Expect 1.5')
        
# --------------------------------------------
class CalcPredii(unittest.TestCase):
    def common_checks(self, ysave, osave):
        self.assertTrue(isinstance(ysave, np.ndarray), msg = 'Expect array')
        self.assertTrue(isinstance(osave, np.ndarray), msg = 'Expect array')
        self.assertEqual(ysave.shape[0], 100, msg = 'Expect 1st dim = 100')
        self.assertEqual(osave.shape[0], 100, msg = 'Expect 1st dim = 100')
        
    def test_calc_predii(self):
        PI, __, testchain, tests2chain, __, sstype, nsample, iisample, datapred = cc_setup()
                
        ysave, osave = PI._calc_credible_and_prediction_ii(testchain = testchain, tests2chain = tests2chain, nrow = 100, ncol = 1, waitbar = False, sstype = 0, test = np.array([True, True]), modelfun = gf.predmodelfun, datapredii = datapred[0])
        self.common_checks(ysave, osave)
        
    def test_calc_predii_with_waitbar(self):
        PI, __, testchain, tests2chain, __, sstype, nsample, iisample, datapred = cc_setup()
        
        PI._PredictionIntervals__wbarstatus = progress_bar(iters = 200)
        
        ysave, osave = PI._calc_credible_and_prediction_ii(testchain = testchain, tests2chain = tests2chain, nrow = 100, ncol = 1, waitbar = True, sstype = 0, test = np.array([True, True]), modelfun = gf.predmodelfun, datapredii = datapred[0])
        self.common_checks(ysave, osave)
        self.assertEqual(PI._PredictionIntervals__wbarstatus.percentage(3), 1.5, msg = 'Expect 1.5')
        
    def test_calc_predii_s2chain_none(self):
        PI, results, testchain, tests2chain, __, sstype, nsample, iisample, datapred = cc_setup(calc_pred_int = False)
        tests2chain = None
        
        with self.assertRaises(TypeError, msg = 'This function should not be called in tests2chain is None'):
            PI._calc_credible_and_prediction_ii(testchain = testchain, tests2chain = tests2chain, nrow = 100, ncol = 1, waitbar = False, sstype = 0, test = np.array([True, True]), modelfun = gf.predmodelfun, datapredii = datapred[0])
                   
    def test_calc_predii_s2chain_tran(self):
        PI, results, testchain, tests2chain, __, sstype, nsample, iisample, datapred = cc_setup(calc_pred_int = False)
        tests2chain = np.random.random_sample(size = (1,100))
        
        with self.assertRaises(SystemExit, msg = 'Unknown structure'):
            PI._calc_credible_and_prediction_ii(testchain = testchain, tests2chain = tests2chain, nrow = 100, ncol = 1, waitbar = False, sstype = 0, test = np.array([True, True]), modelfun = gf.predmodelfun, datapredii = datapred[0])
            
# --------------------------------------------
def pi_setup():
    PI = PredictionIntervals()
    results = gf.setup_pseudo_results()
    DS = gf.basic_data_structure()
    PI.setup_prediction_interval_calculation(results = results, data = DS, modelfunction = gf.predmodelfun)
    return PI, results

class SetupPredictionIntervalCalculation(unittest.TestCase):
    def test_setup_pi_calc(self):
        PI, results = pi_setup()
        self.assertEqual(PI._PredictionIntervals__ndatabatches, 1, msg = 'Expect ndatabatches = 1')
        self.assertEqual(PI.modelfunction, gf.predmodelfun, msg = 'Functions should match')
        self.assertTrue(np.array_equal(PI._PredictionIntervals__s2chain_index, np.array([[0,1]])), msg = str('Arrays should match: {}'.format(PI._PredictionIntervals__s2chain_index)))
        self.assertTrue(np.array_equal(PI._PredictionIntervals__chain, results['chain']), msg = 'Arrays should match')
        
    def test_setup_pi_calc_with_s2chain_none(self):
        PI = PredictionIntervals()
        DS = gf.basic_data_structure()
        results = gf.setup_pseudo_results()
        results['s2chain'] = None
        PI.setup_prediction_interval_calculation(results = results, data = DS, modelfunction = gf.predmodelfun)
        self.assertEqual(PI._PredictionIntervals__ndatabatches, 1, msg = 'Expect ndatabatches = 1')
        self.assertEqual(PI.modelfunction, gf.predmodelfun, msg = 'Functions should match')
        self.assertFalse(hasattr(PI, '_PredictionIntervals__s2chain_index'), msg = 'Expect False')
        self.assertTrue(np.array_equal(PI._PredictionIntervals__chain, results['chain']), msg = 'Arrays should match')

# --------------------------------------------
class CalculateCIForDataSets(unittest.TestCase):
    def test_calc_ci_for_ds(self):
        PI, results = pi_setup()
        PI._assign_features_from_results_structure(results = results)
        lims = np.array([0.025, 0.5, 0.975])
        testchain = np.random.random_sample(size = (100,2))
        ci = PI._calculate_ci_for_data_sets(testchain = testchain, waitbar = False, lims = lims)
        
        self.assertTrue(isinstance(ci, list), msg = 'Expect list return')
        self.assertEqual(ci[0][0].shape[0], 3, msg = 'Expect 1st dim = 3')
        self.assertEqual(ci[0][0].shape[1], 100, msg = 'Expect 2nd dim = 100')
        
        
# --------------------------------------------
class CalculateCIandPIForDataSets(unittest.TestCase):
    def test_calc_ci_and_pi_for_ds(self):
        PI, results = pi_setup()
        PI._assign_features_from_results_structure(results = results)
        lims = np.array([0.025, 0.5, 0.975])
        testchain = np.random.random_sample(size = (100,2))
        s2chain = np.random.random_sample(size = (100,1))
        sstype = 0
        iisample, nsample = PI._define_sample_points(nsample = 400, nsimu = 100)
        ci, pi = PI._calculate_ci_and_pi_for_data_sets(testchain = testchain, s2chain = s2chain, iisample = iisample, waitbar = False, sstype = sstype, lims = lims)
        
        self.assertTrue(isinstance(ci, list), msg = 'Expect list return')
        self.assertEqual(ci[0][0].shape[0], 3, msg = 'Expect 1st dim = 3')
        self.assertEqual(ci[0][0].shape[1], 100, msg = 'Expect 2nd dim = 100')
        self.assertTrue(isinstance(pi, list), msg = 'Expect list return')
        self.assertEqual(pi[0][0].shape[0], 3, msg = 'Expect 1st dim = 3')
        self.assertEqual(pi[0][0].shape[1], 100, msg = 'Expect 2nd dim = 100')
        self.assertEqual(nsample, 100, msg = 'Expect nsample -> 100')
        
# --------------------------------------------
class GeneratePredictionIntervals(unittest.TestCase):
    def common_set_1(self, cint, pint):
        self.assertTrue(isinstance(cint, list), msg = 'Expect list return')
        self.assertEqual(cint[0][0].shape[0], 9, msg = 'Expect 1st dim = 9')
        self.assertEqual(cint[0][0].shape[1], 100, msg = 'Expect 2nd dim = 100')
        self.assertEqual(pint, None, msg = 'Expect None')
        
    def test_generate_credible_intervals(self):
        PI = PredictionIntervals()
        results = gf.setup_pseudo_results()
        results['s2chain'] = None
        DS = gf.basic_data_structure()
        PI.setup_prediction_interval_calculation(results = results, data = DS, modelfunction = gf.predmodelfun)
        PI.generate_prediction_intervals(sstype = None, nsample = 500, calc_pred_int = False, waitbar = False)
        cint = PI.intervals['credible_intervals']
        pint = PI.intervals['prediction_intervals']
        self.common_set_1(cint, pint)
        
    def test_generate_credible_intervals_with_waitbar(self):
        PI = PredictionIntervals()
        results = gf.setup_pseudo_results()
        results['s2chain'] = None
        DS = gf.basic_data_structure()
        PI.setup_prediction_interval_calculation(results = results, data = DS, modelfunction = gf.predmodelfun)
        PI.generate_prediction_intervals(sstype = None, nsample = 500, calc_pred_int = False, waitbar = True)
        cint = PI.intervals['credible_intervals']
        pint = PI.intervals['prediction_intervals']
        self.common_set_1(cint, pint)
        self.assertEqual(PI._PredictionIntervals__wbarstatus.percentage(3), 3.0, msg = 'Expect 3.0')
        
    def test_generate_credible_and_prediction_intervals(self):
        PI, __ = pi_setup()
        PI.generate_prediction_intervals(sstype = 0, nsample = 500, calc_pred_int = True, waitbar = False)
        cint = PI.intervals['credible_intervals']
        pint = PI.intervals['prediction_intervals']
        self.assertTrue(isinstance(cint, list), msg = 'Expect list return')
        self.assertEqual(cint[0][0].shape[0], 3, msg = 'Expect 1st dim = 3')
        self.assertEqual(cint[0][0].shape[1], 100, msg = 'Expect 2nd dim = 100')
        self.assertTrue(isinstance(pint, list), msg = 'Expect list return')
        self.assertEqual(pint[0][0].shape[0], 3, msg = 'Expect 1st dim = 3')
        self.assertEqual(pint[0][0].shape[1], 100, msg = 'Expect 2nd dim = 100')
        
    def test_generate_credible_and_prediction_intervals_with_pi_off(self):
        PI, results = pi_setup()
        PI.generate_prediction_intervals(sstype = 0, nsample = 500, calc_pred_int = False, waitbar = False)
        cint = PI.intervals['credible_intervals']
        pint = PI.intervals['prediction_intervals']
        self.common_set_1(cint, pint)

# --------------------------------------------
class SetupDisplaySettings(unittest.TestCase):
    def test_setup_display_settings(self):
        PI = PredictionIntervals()
        model_display = {'label': 'hello'}
        data_display = {'linewidth': 7}
        interval_display = {'edgecolor': 'b'}
        intd, modd, datd = PI._setup_display_settings(interval_display = interval_display, model_display = model_display, data_display = data_display)
        self.assertEqual(modd['label'], model_display['label'], msg = 'Expect label to match')
        self.assertEqual(intd['edgecolor'], interval_display['edgecolor'], msg = 'Expect edgecolor to match')
        self.assertEqual(datd['linewidth'], data_display['linewidth'], msg = 'Expect linewidth to match')
        
# --------------------------------------------
class PlotPredictionIntervals(unittest.TestCase):
    def common_checks(self, fighandle, axhandle):
        self.assertTrue(isinstance(fighandle, list), msg = 'Expect list return')
        self.assertTrue(isinstance(axhandle, list), msg = 'Expect list return')
        self.assertEqual(fighandle[0].get_label(), 'Batch # 0 | Column # 0', msg = str('Strings should match: {}'.format(fighandle[0].get_label())))

    @classmethod
    def setup_complex_pi(cls):
        PI = PredictionIntervals()
        results = gf.setup_pseudo_results()
        results['s2chain'] = None
        DS = gf.basic_data_structure()
        PI.setup_prediction_interval_calculation(results = results, data = DS, modelfunction = gf.predmodelfun)
        PI.generate_prediction_intervals(sstype = 0, nsample = 500, calc_pred_int = False, waitbar = False)
        return PI
    
    def test_plotting_credible_and_prediction_intervals(self):
        PI, results = pi_setup()
        PI.generate_prediction_intervals(sstype = 0, nsample = 500, calc_pred_int = True, waitbar = False)
        fighandle, axhandle = PI.plot_prediction_intervals(plot_pred_int = True, adddata = False, addlegend = True)
        self.common_checks(fighandle, axhandle)
        self.assertEqual(axhandle[0].get_legend_handles_labels()[1], ['model', '95% PI', '95% CI'], msg = str('Strings should match: {}'.format(axhandle[0].get_legend_handles_labels()[1])))
        plt.close()
        
    def test_plotting_credible_intervals(self):
        PI = self.setup_complex_pi()
        fighandle, axhandle = PI.plot_prediction_intervals(plot_pred_int = False, adddata = False, addlegend = True)
        
        self.common_checks(fighandle, axhandle)
        self.assertEqual(axhandle[0].get_legend_handles_labels()[1], ['model', '99% CI', '95% CI', '90% CI', '50% CI'], msg = str('Strings should match: {}'.format(axhandle[0].get_legend_handles_labels()[1])))
        plt.close()
        
    def test_plotting_credible_intervals_with_data(self):
        PI = self.setup_complex_pi()
        fighandle, axhandle = PI.plot_prediction_intervals(plot_pred_int = False, adddata = True, addlegend = True)
        
        self.common_checks(fighandle, axhandle)
        self.assertEqual(axhandle[0].get_legend_handles_labels()[1], ['model', 'data', '99% CI', '95% CI', '90% CI', '50% CI'], msg = str('Strings should match: {}'.format(axhandle[0].get_legend_handles_labels()[1])))
        plt.close()


# -*- coding: utf-8 -*-

"""
Created on Mon Jan. 15, 2018

Description: This file contains a series of utilities designed to test the
features in the 'PredictionIntervals.py" package of the pymcmcstat module. The
functions tested include:
    - empirical_quantiles(x, p = np.array([0.25, 0.5, 0.75])):

@author: prmiles
"""
from pymcmcstat.settings.ModelParameters import ModelParameters
from pymcmcstat.settings.ModelParameters import format_number_to_str, generate_default_name, check_verbosity, replace_list_elements
from pymcmcstat.settings.ModelParameters import noadapt_display_setting, check_noadaptind, prior_display_setting, less_than_or_equal_to_zero
from pymcmcstat.settings.SimulationOptions import SimulationOptions
import unittest
import numpy as np

# --------------------------
class Add_Model_Parameter_Test(unittest.TestCase):

    def test_does_parameter_assignment_match(self):
        MP = ModelParameters()
        theta0 = 0
        MP.add_model_parameter('aa', theta0)
        self.assertEqual(MP.parameters[0]['theta0'], theta0)
        self.assertEqual(MP.parameters[0]['name'],'aa')
        self.assertEqual(MP.parameters[0]['minimum'], -np.inf)
        self.assertEqual(MP.parameters[0]['maximum'], np.inf)
        self.assertEqual(MP.parameters[0]['prior_mu'],np.zeros([1]))
        self.assertEqual(MP.parameters[0]['prior_sigma'],np.inf)
        self.assertEqual(MP.parameters[0]['sample'],1)
        self.assertEqual(MP.parameters[0]['local'],0)

    def test_does_parameter_assignment_match_with_no_name_or_initial_value(self):
        MP = ModelParameters()
        MP.add_model_parameter(name = None, theta0 = None)
        self.assertEqual(MP.parameters[0]['theta0'], 1.0)
        self.assertEqual(MP.parameters[0]['name'],'$p_{0}$')
        self.assertEqual(MP.parameters[0]['minimum'], -np.inf)
        self.assertEqual(MP.parameters[0]['maximum'], np.inf)
        self.assertEqual(MP.parameters[0]['prior_mu'],np.zeros([1]))
        self.assertEqual(MP.parameters[0]['prior_sigma'],np.inf)
        self.assertEqual(MP.parameters[0]['sample'],1)
        self.assertEqual(MP.parameters[0]['local'],0)

    def test_results_to_params(self):
        MP = ModelParameters()
        MP.add_model_parameter('aa', 0)
        MP._openparameterstructure(nbatch=1)
#        print('parind = {}'.format(MP._parind))
#        print('local = {}'.format(MP._local))
         # define minimal results dictionary
        results = {'parind': MP._parind, 'names': MP._names, 'local': MP._local, 'theta': [1.2]}
        # initialize default options
        SO = SimulationOptions()
        SO.define_simulation_options(verbosity=0)
        MP.display_parameter_settings(verbosity = SO.verbosity, noadaptind = SO.noadaptind)
        MP._results_to_params(results, 1)
        MP._openparameterstructure(nbatch=1)
        MP.display_parameter_settings(verbosity = SO.verbosity, noadaptind = SO.noadaptind)
        self.assertEqual(MP.parameters[0]['theta0'], results['theta'][0])
                
# --------------------------
MP = ModelParameters()
class LessThanOrEqualToZero(unittest.TestCase):
    
    def test_x_lt_0(self):
        self.assertTrue(less_than_or_equal_to_zero(x = -1))
        
    def test_x_eq_0(self):
        self.assertTrue(less_than_or_equal_to_zero(x = 0))
        
    def test_x_gt_0(self):
        self.assertFalse(less_than_or_equal_to_zero(x = 1))

# --------------------------
class ReplaceListElements(unittest.TestCase):
    
    def test_list_of_0s(self):
        x = [0, 0, 0]
        self.assertEqual(replace_list_elements(x = x, testfunction = less_than_or_equal_to_zero, value = 10), [10, 10, 10])
        
    def test_list_of_neg(self):
        x = [-1, -1, -1]
        self.assertEqual(replace_list_elements(x = x, testfunction = less_than_or_equal_to_zero, value = 10), [10, 10, 10])
        
    def test_list_of_pos(self):
        x = [1, 1, 1]
        self.assertEqual(replace_list_elements(x = x, testfunction = less_than_or_equal_to_zero, value = 10), x)
        
    def test_list_of_mixed(self):
        x = [1, 0, -1]
        self.assertEqual(replace_list_elements(x = x, testfunction = less_than_or_equal_to_zero, value = 10), [1, 10, 10])
        
# --------------------------
class DisplayParametersSettings(unittest.TestCase):
    
    def test_parameter_display_set_1(self):
        MP.add_model_parameter(name = 'm', theta0 = 2., minimum = -10, maximum = np.inf, sample = 1)
        MP.add_model_parameter(name = 'b', theta0 = -5., minimum = -10, maximum = 100, sample = 0)
        MP.add_model_parameter(name = 'b2', theta0 = -5.3e6, minimum = -1e7, maximum = 1e6, sample = 1)
        MP._openparameterstructure(nbatch = 1)
        
        MP.display_parameter_settings(verbosity = 1, noadaptind = None)
        
    def test_parameter_display_set_2(self):
        MP.add_model_parameter(name = 'm', theta0 = 2., minimum = -10, maximum = np.inf, sample = 1)
        MP.add_model_parameter(name = 'b', theta0 = -5., minimum = -10, maximum = 100, sample = 0)
        MP.add_model_parameter(name = 'b2', theta0 = -5.3e6, minimum = -1e7, maximum = 1e6, sample = 1)
        MP._openparameterstructure(nbatch = 1)
        
        MP.display_parameter_settings(verbosity = None, noadaptind = None)

# --------------------------
class NoadaptindDisplaySetting(unittest.TestCase):
    
    def test_noadaptind_is_empty(self):
        self.assertEqual(noadapt_display_setting(ii = 1, noadaptind = []), '', msg = 'Default is blank string')
        self.assertEqual(noadapt_display_setting(ii = 2, noadaptind = []), '', msg = 'Default is blank string')
        
    def test_noadaptind_is_not_empty(self):
        self.assertEqual(noadapt_display_setting(ii = 1, noadaptind = [2]), '', msg = 'Default is blank string')
        self.assertEqual(noadapt_display_setting(ii = 2, noadaptind = [2]), ' (*)', msg = 'Default is blank string')

# --------------------------
class PriorDisplaySetting(unittest.TestCase):
    
    def test_x_is_inf(self):
        self.assertEqual(prior_display_setting(x = np.inf), '', msg = 'Blank string if infinity.')
        
    def test_x_is_not_inf(self):
        self.assertEqual(prior_display_setting(x = 2.0), '^2', msg = 'Raised to the second power if not infinity.')
        
# --------------------------        
class CheckNoadaptind(unittest.TestCase):
    
    def test_noadaptind_is_none(self):
        self.assertEqual(check_noadaptind(noadaptind = None), [], msg = 'Returns empty list.')
        
    def test_noadaptind_is_not_none(self):
        self.assertEqual(check_noadaptind(noadaptind = [1]), [1], msg = 'Returns input.')
        
# --------------------------        
class Verbosity(unittest.TestCase):
    
    def test_verbosity_is_none(self):
        self.assertEqual(check_verbosity(verbosity = None), 0, msg = 'Returns 0.')
        
    def test_verbosity_is_not_none(self):
        self.assertEqual(check_verbosity(verbosity = 1), 1, msg = 'Returns input.')
        
# --------------------------        
class ParameterLimits(unittest.TestCase):
    
    def test_initial_parameter_outside(self):
        MP = ModelParameters()
        MP.add_model_parameter('aa', 0, minimum = -10., maximum = 10)
        MP.add_model_parameter('bb', 11, minimum = -10., maximum = 10)
        MP._openparameterstructure(nbatch=1)
        with self.assertRaises(SystemExit, msg = 'Initial value outside of bounds.'):
            MP._check_initial_values_wrt_parameter_limits()
            
    def test_initial_parameter_okay(self):
        MP = ModelParameters()
        MP.add_model_parameter('aa', 0, minimum = -10., maximum = 10)
        MP.add_model_parameter('bb', 9, minimum = -10., maximum = 10)
        MP._openparameterstructure(nbatch=1)
        self.assertTrue(MP._check_initial_values_wrt_parameter_limits())
        
# --------------------------        
class GenerateDefaultName(unittest.TestCase):
    
    def test_default_name_generation(self):
        self.assertEqual(generate_default_name(nparam = 0), '$p_{0}$', msg = '0 based naming convention.')
        self.assertEqual(generate_default_name(nparam = 3), '$p_{3}$', msg = '0 based naming convention.')
        self.assertEqual(generate_default_name(nparam = 2), '$p_{2}$', msg = '0 based naming convention.')
        self.assertEqual(generate_default_name(nparam = 10), '$p_{10}$', msg = '0 based naming convention.')

# --------------------------
class FormatNumberToStr(unittest.TestCase):
    def test_format_number_to_str(self):
        self.assertEqual(str('{:9.2f}'.format(1.0)), format_number_to_str(1.0), msg = str('Exect: {:9.2f}'.format(1.0)))
        self.assertEqual(str('{:9.2e}'.format(1.0e4)), format_number_to_str(1.0e4), msg = str('Exect: {:9.2f}'.format(1.0e4)))
        self.assertEqual(str('{:9.2e}'.format(1.0e-2)), format_number_to_str(1.0e-2), msg = str('Exect: {:9.2f}'.format(1.0e-2)))


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 18 09:10:21 2018

Description: Prior function

@author: prmiles
"""
# Import required packages
import numpy as np

class PriorFunction:
    '''
    Prior distribution functions.

    Attributes:
        * :meth:`default_priorfun`
        * :meth:`evaluate_prior`
    '''
    def __init__(self, priorfun = None, mu = None, sigma = None):

        self.mu = mu
        self.sigma = sigma

        # Setup prior function and evaluate
        if priorfun is None:
            priorfun = self.default_priorfun

        self.priorfun = priorfun # function handle
        
    @classmethod
    def default_priorfun(cls, theta, mu, sigma):
        '''
        Default prior function.

        .. math::

            \\pi_0(q) = \sum_{i=1}^N \\Big(\\frac{\\theta_i - \\mu_i}{\\sigma_i^2}\\Big)^2

        Args:
            * **theta** (:class:`~numpy.ndarray`): Current parameter values.
            * **mu** (:class:`~numpy.ndarray`): Prior mean.
            * **sigma** (:class:`~numpy.ndarray`): Prior standard deviation.
        '''
        # consider converting everything to numpy array - should allow for optimized performance
        n = len(theta)
        
        if mu is None:
            mu = np.zeros([n,1])
        
        if sigma is None:
            sigma = np.ones([n,1])
        
        pf = np.zeros(1)
        for ii in range(n):
            pf = pf + ((theta[ii]-mu[ii])*(sigma[ii]**(-1)))**2

        return pf
        
    def evaluate_prior(self, theta):
        '''
        Evaluate the prior function.

        Args:
            * **theta** (:class:`~numpy.ndarray`): Current parameter values.
        '''
        return self.priorfun(theta, self.mu, self.sigma)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 30 09:47:02 2018

@author: prmiles
"""
from pymcmcstat.samplers.SamplingMethods import SamplingMethods
from pymcmcstat.samplers.Metropolis import Metropolis
from pymcmcstat.samplers.DelayedRejection import DelayedRejection
from pymcmcstat.samplers.Adaptation import Adaptation

import unittest

# -------------------
class CreateSamplingMethods(unittest.TestCase):
    def test_create_sampling_methods(self):
        SM = SamplingMethods()
        self.assertTrue(isinstance(SM.metropolis, Metropolis), msg = 'Metropolis Class')
        self.assertTrue(isinstance(SM.delayed_rejection, DelayedRejection), msg = 'Delayed Rejection Class')
        self.assertTrue(isinstance(SM.adaptation, Adaptation), msg = 'Adaptation Class')


# -*- coding: utf-8 -*-

"""
Created on Mon Jan. 15, 2018

Description: This file contains a series of utilities designed to test the
features in the 'ParameterSet.py" class of the pymcmcstat module.

@author: prmiles
"""
from pymcmcstat.structures.ParameterSet import ParameterSet
import test.general_functions as gf
import unittest

# --------------------------
# ParameterSet
# --------------------------
class Initialize_Parameter_Set(unittest.TestCase):

    def test_PS_default_match(self):
        PS = ParameterSet()
        PSD = PS.__dict__
        for (k,v) in PSD.items():
            self.assertEqual(v, None, msg = str('Default {} is None'.format(k)))

    def test_PS_set_theta(self):
        x = 1.2
        key = 'theta'
        PS = ParameterSet(theta = x)
        PSD = PS.__dict__
        self.assertEqual(PSD[key], x, msg = str('Expected {} = {}'.format(key, x)))
        PSD = gf.removekey(PSD, key)
        for (k,v) in PSD.items():
            self.assertEqual(v, None, msg = str('Default {} is None'.format(k)))
        
    def test_PS_set_ss(self):
        x = 1.2
        key = 'ss'
        PS = ParameterSet(ss = x)
        PSD = PS.__dict__
        self.assertEqual(PSD[key], x, msg = str('Expected {} = {}'.format(key, x)))
        PSD = gf.removekey(PSD, key)
        for (k,v) in PSD.items():
            self.assertEqual(v, None, msg = str('Default {} is None'.format(k)))
        
    def test_PS_set_prior(self):
        x = 1.2
        key = 'prior'
        PS = ParameterSet(prior = x)
        PSD = PS.__dict__
        self.assertEqual(PSD[key], x, msg = str('Expected {} = {}'.format(key, x)))
        PSD = gf.removekey(PSD, key)
        for (k,v) in PSD.items():
            self.assertEqual(v, None, msg = str('Default {} is None'.format(k)))
            
    def test_PS_set_sigma2(self):
        x = 1.2
        key = 'sigma2'
        PS = ParameterSet(sigma2 = x)
        PSD = PS.__dict__
        self.assertEqual(PSD[key], x, msg = str('Expected {} = {}'.format(key, x)))
        PSD = gf.removekey(PSD, key)
        for (k,v) in PSD.items():
            self.assertEqual(v, None, msg = str('Default {} is None'.format(k)))
            
    def test_PS_set_alpha(self):
        x = 1.2
        key = 'alpha'
        PS = ParameterSet(alpha = x)
        PSD = PS.__dict__
        self.assertEqual(PSD[key], x, msg = str('Expected {} = {}'.format(key, x)))
        PSD = gf.removekey(PSD, key)
        for (k,v) in PSD.items():
            self.assertEqual(v, None, msg = str('Default {} is None'.format(k)))
            
    def test_PS_set_theta_and_prior(self):
        x = 1.2
        keys = ['theta', 'prior']
        PS = ParameterSet(theta = x, prior = x)
        PSD = PS.__dict__
        for key in keys:
            self.assertEqual(PSD[key], x, msg = str('Expected {} = {}'.format(key, x)))
        for key in keys:
            PSD = gf.removekey(PSD, key)
        for (k,v) in PSD.items():
            self.assertEqual(v, None, msg = str('Default {} is None'.format(k)))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 31 11:52:58 2018

@author: prmiles
"""

from pymcmcstat.settings.SimulationOptions import SimulationOptions
from pymcmcstat.settings.ModelSettings import ModelSettings
import unittest

def setup_options(**kwargs):
    SO = SimulationOptions()
    SO.define_simulation_options(**kwargs)
    return SO

MS = ModelSettings()
MS.define_model_settings()

# --------------------------
class DisplaySimulationOptions(unittest.TestCase):
    
    def test_print_these_none(self):
        SO = setup_options()
        print_these = SO.display_simulation_options(print_these = None)
        self.assertEqual(print_these, ['nsimu', 'adaptint', 'ntry', 'method', 'printint', 'lastadapt', 'drscale', 'qcov'], msg = 'Default print keys')
        
    def test_print_these_not_none(self):
        SO = setup_options()
        print_these = SO.display_simulation_options(print_these = ['nsimu'])
        self.assertEqual(print_these, ['nsimu'], msg = 'Specified print keys')
        
# --------------------------
class CheckDependentOptions(unittest.TestCase):
    
    def test_dodram(self):
        SO = setup_options(ntry = 1)
        SO._check_dependent_simulation_options(model = MS)
        self.assertEqual(SO.dodram, 0, msg = 'DRAM turned off because ntry <= 1')
        SO = setup_options(ntry = 2)
        SO._check_dependent_simulation_options(model = MS)
        self.assertEqual(SO.dodram, 1, msg = 'DRAM turned on because ntry > 1')
        
    def test_lastadapt(self):
        SO = setup_options(lastadapt = 0)
        SO._check_dependent_simulation_options(model = MS)
        self.assertEqual(SO.lastadapt, SO.nsimu, msg = 'lastadapt set to nsimu')
        SO.define_simulation_options(lastadapt = 10)
        SO._check_dependent_simulation_options(model = MS)
        self.assertEqual(SO.lastadapt, 10, msg = 'lastadapt unchanged')
        
    def test_printint(self):
        SO = setup_options(printint = None)
        SO._check_dependent_simulation_options(model = MS)
        self.assertEqual(SO.printint, max(100,min(1000,SO.adaptint)), msg = 'printint was updated')
        
    def test_updatesigma(self):
        SO = setup_options(updatesigma = False)
        MS.N0 = [1.]
        SO._check_dependent_simulation_options(model = MS)
        self.assertTrue(SO.updatesigma, msg = 'updatesigma turned on because N0 not empty')
        
# --------------------------
class DefineOptions(unittest.TestCase):
    
    def test_define_options_method_assignment(self):
        methods = ['mh', 'dr', 'am', 'dram']
        for method in methods:
            SO = setup_options(method = method)
            self.assertEqual(SO.method, method, msg = str('Method set to {}'.format(method)))
            
    def test_ntry_assigment(self):
        SO = setup_options(ntry = 3)
        self.assertEqual(SO.ntry, 3, msg = 'ntry set to 3')
        
    def test_adascale_assigment(self):
        SO = setup_options(adascale = 3)
        self.assertEqual(SO.adascale, 3, msg = 'adascale set to 3')
        
    def test_label_assigment(self):
        SO = setup_options(label = '3')
        self.assertEqual(SO.label, '3', msg = 'label set to 3')
        
    def test_savedir_assigment(self):
        SO = setup_options(savedir = '3')
        self.assertEqual(SO.savedir, '3', msg = 'savedir set to 3')


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 17, 2018

@author: prmiles

Description: This module is intended to be the main class from which to run these
Markov Chain Monte Carlo type simulations.  The user will create an MCMC object,
initialize data, simulation options, model settings and parameters.
"""

# import required packages
#import sys
import os
import time
import numpy as np
import datetime
import sys

from .settings.DataStructure import DataStructure
from .settings.ModelSettings import ModelSettings
from .settings.ModelParameters import ModelParameters
from .settings.SimulationOptions import SimulationOptions

from .procedures.CovarianceProcedures import CovarianceProcedures
from .procedures.SumOfSquares import SumOfSquares
from .procedures.PriorFunction import PriorFunction
from .procedures.ErrorVarianceEstimator import ErrorVarianceEstimator

from .structures.ParameterSet import ParameterSet
from .structures.ResultsStructure import ResultsStructure

from .samplers.SamplingMethods import SamplingMethods

from .plotting import MCMCPlotting
from .plotting.PredictionIntervals import PredictionIntervals

from .chain import ChainStatistics
from .chain import ChainProcessing

from .utilities.progressbar import progress_bar
from .utilities.general import message
# --------------------------------------------------------
class MCMC:
    '''
    Markov Chain Monte Carlo (MCMC) simulation object.
    
    This class type is the primary feature of this Python package.  All simulations
    are run through this class type, and for the most part the user will interact
    with an object of this type.  The class initialization provides the option for 
    setting the random seed, which can be very useful for testing the functionality 
    of the code.  It was found that setting the random seed at object initialization
    was the simplest interface.
    
    Args:
        * **rngseed** (:py:class:`float`): Seed for numpy's random number generator.

    Attributes:
        * :meth:`~run_simulation`
        * :meth:`~display_current_mcmc_settings`
        * **data** (:class:`~.DataStructure`): MCMC data structure.
        * **simulation_options** (:class:`~.SimulationOptions`): MCMC simulation options.
        * **model_settings** (:class:`~.ModelSettings`): MCMC model settings.
        * **parameters** (:class:`~.ModelParameters`): MCMC model parameters.
    '''
    def __init__(self, rngseed = None):
        # public variables
        self.data = DataStructure()
        self.model_settings = ModelSettings()
        self.simulation_options = SimulationOptions()
        self.parameters = ModelParameters()
        # private variables
        self._error_variance = ErrorVarianceEstimator()
        self._covariance = CovarianceProcedures()
        self._sampling_methods = SamplingMethods()
        self._mcmc_status = False
        np.random.seed(seed = rngseed)
    # --------------------------------------------------------
    def run_simulation(self, use_previous_results = False):
        '''
        Run MCMC Simulation
        
        .. note::

            This is the method called by the user to run the simulation.  The user
            must specify a data structure, setup simulation options, and define
            the model settings and parameters before calling this method.

        Args:
            * **use_previous_results** (:py:class:`bool`): Flag to indicate whether simulation is being restarted.
        '''
        start_time = time.time()
        
        self.__setup_simulator(use_previous_results = use_previous_results)
        
        # ---------------------
        # setup progress bar
        if self.simulation_options.waitbar:
            self.__wbarstatus = progress_bar(iters = int(self.simulation_options.nsimu))
            
        # ---------------------
        # displacy current settings
        if self.simulation_options.verbosity >= 2:
            self.display_current_mcmc_settings()
            
        # ---------------------
        # Execute main simulator
        self.__execute_simulator()
        
        end_time = time.time()
        self.__simulation_time = end_time - start_time
        # --------------------
        # Generate Results
        self.__generate_simulation_results()
        if self.simulation_options.save_to_json == True:
            if self.simulation_results.basic == True: # check that results structure has been created
                self.simulation_results.export_simulation_results_to_json_file(results = self.simulation_results.results)
        self.mcmcplot = MCMCPlotting.Plot()
        self.PI = PredictionIntervals()
        self.chainstats = ChainStatistics.chainstats
        self._mcmc_status = True # simulation has been performed
    # --------------------------------------------------------
    def __setup_simulator(self, use_previous_results):
        '''
        Setup simulator.
        
        If previous results exist, then chain arrays will be expanded.  Otherwise,
        they will be initialized.
        
        If `use_previous_results` is `False`, one can still restart a simulation
        by specifying a `json_restart_file` in the :class:`~.SimulationOptions`.

        Args:
            * **use_previous_results** (:py:class:`bool`): Flag to indicate whether simulation is being restarted.
        '''
        if use_previous_results == True:
            if self._mcmc_status == True:
                self.parameters._results_to_params(self.simulation_results.results, 1)
                self._initialize_simulation()
                self.__expand_chains(nsimu = self.simulation_options.nsimu, npar = self.parameters.npar, nsos = self.model_settings.nsos, updatesigma=self.simulation_options.updatesigma)
            else:
                sys.exit('No previous results found.  Set ''use_previous_results'' to ''False''')
        else:
            if self.simulation_options.json_restart_file is not None:
                RS = ResultsStructure()
                res = RS.load_json_object(self.simulation_options.json_restart_file)
                self.parameters._results_to_params(res, 1)
                self.simulation_options.qcov = np.array(res['qcov'])
                
            self.__chain_index = 0 # start index at zero
            self._initialize_simulation()
            self.__initialize_chains(chainind = self.__chain_index, nsimu = self.simulation_options.nsimu, npar = self.parameters.npar, nsos = self.model_settings.nsos, updatesigma=self.simulation_options.updatesigma, sigma2 = self.model_settings.sigma2)
    # --------------------------------------------------------
    def _initialize_simulation(self):
        '''
        Initialize all dependent settings for simulation.
        '''
        # ---------------------------------
        # check dependent parameters
        self.simulation_options._check_dependent_simulation_options(self.model_settings)
        self.model_settings._check_dependent_model_settings(self.data, self.simulation_options)
        # open and parse the parameter structure
        self.parameters._openparameterstructure(self.model_settings.nbatch)
        # check initial parameter values are inside range
        self.parameters._check_initial_values_wrt_parameter_limits()
        # add check that prior standard deviation > 0
        self.parameters._check_prior_sigma(self.simulation_options.verbosity)
        # display parameter settings
        self.parameters.display_parameter_settings(self.simulation_options.verbosity, self.simulation_options.noadaptind)
        # setup covariance matrix and initial Cholesky decomposition
        self._covariance._initialize_covariance_settings(self.parameters, self.simulation_options)
        # ---------------------
        # define sum-of-squares object
        self.__sos_object = SumOfSquares(self.model_settings, self.data, self.parameters)
        # ---------------------
        # define prior object
        self.__prior_object = PriorFunction(priorfun = self.model_settings.prior_function, mu = self.parameters._thetamu, sigma = self.parameters._thetasigma)
        # ---------------------
        # Define initial parameter set
        self.__initial_set = ParameterSet(theta = self.parameters._initial_value[self.parameters._parind[:]])
        # calculate sos with initial parameter set
        self.__initial_set.ss = self.__sos_object.evaluate_sos_function(self.__initial_set.theta)
        nsos = len(self.__initial_set.ss)
        # evaluate prior with initial parameter set
        self.__initial_set.prior = self.__prior_object.evaluate_prior(self.__initial_set.theta)
        # add initial error variance to initial parameter set
        self.__initial_set.sigma2 = self.model_settings.sigma2
        # recheck certain values in model settings that are dependent on the output of the sos function
        self.model_settings._check_dependent_model_settings_wrt_nsos(nsos)
        # ---------------------
        # Update variables covariance adaptation
        self._covariance._update_covariance_settings(self.__initial_set.theta)
        
        if self.simulation_options.ntry > 1:
            self._sampling_methods.delayed_rejection._initialize_dr_metrics(self.simulation_options)
    # --------------------------------------------------------
    def __initialize_chains(self, chainind, nsimu, npar, nsos, updatesigma, sigma2):
        '''
        Initialize chains

        Args:
            * **chainind** (:py:class:`int`): Where to store initial parameter value
            * **nsimu** (:py:class:`int`): Number of parameter samples to simulate.  Default is 1e4.
            * **npar** (:py:class:`int`): Number of parameters being sampled.
            * **nsos** (:py:class:`int`): Length of output from sum-of-squares function
            * **updatesigma** (:py:class:`bool`): Flag for updating measurement error variance. Default is 0 -> off (1 -> on).
            * **sigma2** (:class:`numpy.ndarray`): Initial error observations.
        '''
        # Initialize chain, error variance, and SS
        self.__chain = np.zeros([nsimu, npar])
        self.__sschain = np.zeros([nsimu, nsos])
        # Save initialized values to chain, s2chain, sschain
        self.__chain[chainind,:] = self.__initial_set.theta
        self.__sschain[chainind,:] = self.__initial_set.ss
        
        if updatesigma:
            self.__s2chain = np.zeros([nsimu, nsos])
            self.__s2chain[chainind,:] = sigma2
        else:
            self.__s2chain = None
    # --------------------------------------------------------    
    def __expand_chains(self, nsimu, npar, nsos, updatesigma):
        '''
        Expand chains for extended simulation

        Args:
            * **nsimu** (:py:class:`int`): Number of parameter samples to simulate.  Default is 1e4.
            * **npar** (:py:class:`int`): Number of parameters being sampled.
            * **nsos** (:py:class:`int`): Length of output from sum-of-squares function
            * **updatesigma** (:py:class:`bool`): Flag for updating measurement error variance. Default is 0 -> off (1 -> on).
        '''
        # continuing simulation, so we must expand storage arrays
        zero_chain = np.zeros([nsimu-1, npar])
        zero_sschain = np.zeros([nsimu-1, nsos])
        # Concatenate with previous chains
        self.__chain = np.concatenate((self.__chain, zero_chain), axis = 0)
        self.__sschain = np.concatenate((self.__sschain, zero_sschain), axis = 0)
        if updatesigma:
            zero_s2chain = np.zeros([nsimu-1, nsos])
            self.__s2chain = np.concatenate((self.__s2chain, zero_s2chain), axis = 0)
        else:
            self.__s2chain = None
    # --------------------------------------------------------
    def __execute_simulator(self):
        '''
        Execute MCMC simulation.
        
        Runs through `nsimu` simulations using `method` defined by user.
        '''
        iiadapt = 0 # adaptation counter
        iiprint = 0 # print counter
        savecount = 0 # save counter
        lastbin = 0 # initialize bin counter
        nsimu = self.simulation_options.nsimu
                
        self.__rejected = {'total': 0, 'in_adaptation_interval': 0, 'outside_bounds': 0}
        self.__old_set = self.__initial_set
        
        for isimu in range(1, nsimu): # simulation loop
            # update indexing
            iiadapt += 1 # local adaptation index
            iiprint += 1 # local print index
            savecount += 1 # counter for saving to bin files
            self.__chain_index += 1
            # progress bar
            if self.simulation_options.waitbar:
                self.__wbarstatus.update(isimu)
                
            message(self.simulation_options.verbosity, 100, str('i: {:d}/{:d}\n'.format(isimu,nsimu)));

            # METROPOLIS
            accept, new_set, outbound, npar_sample_from_normal = self._sampling_methods.metropolis.run_metropolis_step(
                    old_set = self.__old_set, parameters = self.parameters, R = self._covariance._R,
                    prior_object = self.__prior_object, sos_object = self.__sos_object)

            # DELAYED REJECTION
            if self.simulation_options.ntry > 1 and accept == 0:
                accept, new_set, outbound = self._sampling_methods.delayed_rejection.run_delayed_rejection(
                        old_set = self.__old_set, new_set = new_set, RDR = self._covariance._RDR, ntry = self.simulation_options.ntry,
                        parameters = self.parameters, invR = self._covariance._invR,
                        sosobj = self.__sos_object, priorobj = self.__prior_object)

            # UPDATE CHAIN & SUM-OF-SQUARES CHAIN
            self.__update_chain(accept = accept, new_set = new_set, outsidebounds = outbound)
            self.__sschain[self.__chain_index,:] = self.__old_set.ss

            # PRINT REJECTION STATISTICS
            if self.simulation_options.printint and iiprint == self.simulation_options.printint:
                print_rejection_statistics(rejected = self.__rejected, isimu = isimu, iiadapt = iiadapt, verbosity = self.simulation_options.verbosity)
                iiprint = 0 # reset print counter

            # ADAPTATION
            if self.simulation_options.adaptint > 0 and iiadapt == self.simulation_options.adaptint:
                self._covariance = self._sampling_methods.adaptation.run_adaptation(
                        covariance = self._covariance, options = self.simulation_options,
                        isimu = isimu, iiadapt = iiadapt, rejected = self.__rejected,
                        chain = self.__chain, chainind = self.__chain_index, u = npar_sample_from_normal,
                        npar = self.parameters.npar, alpha = new_set.alpha)
                
                iiadapt = 0 # reset local adaptation index
                self.__rejected['in_adaptation_interval'] = 0 # reset local rejection index
            
            # UPDATE ERROR VARIANCE
            if self.simulation_options.updatesigma:
                sigma2 = self._error_variance.update_error_variance(self.__old_set.ss, self.model_settings)
                self.__s2chain[self.__chain_index,:] = sigma2
                self.__old_set.sigma2 = sigma2
                
            # SAVE TO LOG FILE
            if savecount == self.simulation_options.savesize:
                savecount, lastbin = self.__save_to_log_file(start = isimu - self.simulation_options.savesize, end = isimu)
       
        # SAVE REMAINING ELEMENTS TO BIN FILE
        self.__save_to_log_file(start = lastbin, end = isimu + 1)
           
        # update value to end value
        self.parameters._value[self.parameters._parind] = self.__old_set.theta
    # ------------------------------------------------
    def __generate_simulation_results(self):
        '''
        Generate simulation results dictionary.
        
        - Basic
        - Delayed rejection stats (if applicable)
        - Simulation options
        - Model settings
        - Sampling chain
        - Observation error chain
        - Sum-of-squares chain
        '''
        # --------------------------------------------
        # BUILD RESULTS OBJECT
        self.simulation_results = ResultsStructure() # inititialize
        self.simulation_results.add_basic(nsimu = self.simulation_options.nsimu, covariance = self._covariance, parameters = self.parameters, rejected = self.__rejected, simutime = self.__simulation_time, theta = self.parameters._value)

        if self.simulation_options.ntry > 1:
            self.simulation_results.add_dram(drscale = self.simulation_options.drscale, RDR = self._covariance._RDR, total_rejected = self.__rejected['total'], drsettings = self._sampling_methods.delayed_rejection)
        
        self.simulation_results.add_options(options = self.simulation_options)
        self.simulation_results.add_model(model = self.model_settings)
        
        # add chain, s2chain, and sschain
        self.simulation_results.add_chain(chain = self.__chain)
        self.simulation_results.add_s2chain(s2chain = self.__s2chain)
        self.simulation_results.add_sschain(sschain = self.__sschain)
    # ------------------------------------------------
    def __save_to_log_file(self, start, end):
        '''
        Save to log files
        
        Args:
            * **start** (:py:class:`int`): Start index of chain block to save
            * **end** (:py:class:`int`): End index of chain block to save
            
        Returns:
            * **savecount** (:py:class:`int`): Reset save counter
            * **lastbin** (:py:class:`int`): Last index saved
        '''
        if self.simulation_options.save_to_bin is True:
            self.__save_chains_to_bin(start, end)
        if self.simulation_options.save_to_txt is True:
            self.__save_chains_to_txt(start, end)
            
        # reset counter
        savecount = 0
        lastbin = end
        return savecount, lastbin
    # --------------------------------------------------------
    def __save_chains_to_bin(self, start, end):
        '''
        Save chain segment to binary file

        Args:
            * **start** (:py:class:`int`): Starting index of chain to save to file
            * **end** (:py:class:`int`): Ending index of chain to save to file

        If you specify a `savesize` of 100, then every 100 simulations the last 100
        chain sets will be appended to the file.  That is to say, if you are on
        simulation 1000, the chain elements 900-999 will be appended to the file.
        '''
        savedir = self.simulation_options.savedir
        ChainProcessing._check_directory(savedir)
        
        chainfile, s2chainfile, sschainfile, covchainfile = ChainProcessing._create_path_with_extension_for_all_logs(self.simulation_options, extension = 'h5')
        
        binlogfile = os.path.join(savedir, 'binlogfile.txt')
        binstr = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        ChainProcessing._add_to_log(binlogfile, str('{}\t{}\t{}\n'.format(binstr, start, end-1)))

        # define set name based in start/end
        datasetname = str('{}_{}_{}'.format('nsimu',start,end-1))
        
        ChainProcessing._save_to_bin_file(chainfile, datasetname = datasetname, mtx = self.__chain[start:end,:])
        ChainProcessing._save_to_bin_file(sschainfile, datasetname = datasetname, mtx = self.__sschain[start:end,:])
        ChainProcessing._save_to_bin_file(covchainfile, datasetname = datasetname, mtx = np.dot(self._covariance._R.transpose(),self._covariance._R))
        
        if self.simulation_options.updatesigma == 1:
            ChainProcessing._save_to_bin_file(s2chainfile, datasetname = datasetname, mtx = self.__s2chain[start:end,:])
    # --------------------------------------------------------
    def __save_chains_to_txt(self, start, end):
        '''
        Save chain segment to text file

        Args:
            * **start** (:py:class:`int`): Starting index of chain to save to file
            * **end** (:py:class:`int`): Ending index of chain to save to file

        If you specify a `savesize` of 100, then every 100 simulations the last 100
        chain sets will be appended to the file.  That is to say, if you are on
        simulation 1000, the chain elements 900-999 will be appended to the file.
        '''
        savedir = self.simulation_options.savedir
        ChainProcessing._check_directory(savedir)
        
        chainfile, s2chainfile, sschainfile, covchainfile = ChainProcessing._create_path_with_extension_for_all_logs(self.simulation_options, extension = 'txt')
       
        txtlogfile = os.path.join(savedir, 'txtlogfile.txt')
        txtstr = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        ChainProcessing._add_to_log(txtlogfile, str('{}\t{}\t{}\n'.format(txtstr, start, end-1)))
        
        ChainProcessing._save_to_txt_file(chainfile, self.__chain[start:end,:])
        ChainProcessing._save_to_txt_file(sschainfile, self.__sschain[start:end,:])
        ChainProcessing._save_to_txt_file(covchainfile, np.dot(self._covariance._R.transpose(),self._covariance._R))
        
        if self.simulation_options.updatesigma == 1:
            ChainProcessing._save_to_txt_file(s2chainfile, self.__s2chain[start:end,:])
    
    # --------------------------------------------------------
    def __update_chain(self, accept, new_set, outsidebounds):
        '''
        Update chain

        Args:
            * **accept** (:py:class:`str`): Flag to indicate whether :math:`q^*` is accepted or rejected
            * **new_set** (:class:`~.ParameterSet`): Features of :math:`q^*`
            * **outsidebounds** (:py:class:`bool`): Flag to indicate whether rejection occured due to sampling outside limits
        '''
        if accept:
            # accept
            self.__chain[self.__chain_index,:] = new_set.theta
            self.__old_set = new_set
        else:
            # reject
            self.__chain[self.__chain_index,:] = self.__old_set.theta
            self.__update_rejected(outsidebounds = outsidebounds)
    # --------------------------------------------------------
    def __update_rejected(self, outsidebounds):
        '''
        Update rejection counters

        Args:
            * **outsidebounds** (:py:class:`bool`): Flag to indicate whether rejection occured due to sampling outside limits
        '''
        self.__rejected['total'] += 1
        self.__rejected['in_adaptation_interval'] += 1
        if outsidebounds:
            self.__rejected['outside_bounds'] += 1
    # --------------------------------------------------------        
    def display_current_mcmc_settings(self):
        '''
        Display model settings, simulation options, and current covariance values.
        
        Example display:
            
        ::
            
            model settings:
                sos_function = <function test_ssfun at 0x1c13c5d400>
                model_function = None
                sigma2 = [1.]
                N = [100.]
                N0 = [0.]
                S20 = [1.]
                nsos = 1
                nbatch = 1
            simulation options:
                nsimu = 5000
                adaptint = 100
                ntry = 2
                method = dram
                printint = 100
                lastadapt = 5000
                drscale = [5. 4. 3.]
                qcov = None
            covariance:
                qcov = [[0.01   0.    ]
                [0.     0.0625]]
                R = [[0.1  0.  ]
                [0.   0.25]]
                RDR = [array([[0.1 , 0.  ],
               [0.  , 0.25]]), array([[0.02, 0.  ],
               [0.  , 0.05]])]
                invR = [array([[10.,  0.],
               [ 0.,  4.]]), array([[50.,  0.],
               [ 0., 20.]])]
                last_index_since_adaptation = 0
                covchain = None
        '''
        self.model_settings.display_model_settings()
        self.simulation_options.display_simulation_options()
        self._covariance.display_covariance_settings()
# --------------------------------------------------------
def print_rejection_statistics(rejected, isimu, iiadapt, verbosity):
    '''
    Print Rejection Statistics.
    
    Threshold for printing is verbosity greater than or equal to 2.  If the rejection
    counters are as follows:
        
        - `total`: 144
        - `in_adaptation_interval`: 92
        - `outside_bounds`: 0
        
    Then we would expect the following display at the 200th simulation with an adaptation
    interval of 100.

    ::

        i:200 (72.0,92.0,0.0)

    Args:
        * **isimu** (:py:class:`int`): Simulation counter
        * **iiadapt** (:py:class:`int`): Adaptation counter
        * **verbosity** (:py:class:`int`): Verbosity of display output.
    '''
    message(verbosity, 2, str('i:{} ({},{},{})\n'.format(
            isimu, rejected['total']*isimu**(-1)*100, rejected['in_adaptation_interval']*iiadapt**(-1)*100,
            rejected['outside_bounds']*isimu**(-1)*100)))